import inspect
import logging
from abc import ABC, abstractmethod
from datetime import datetime, timedelta

import pandas as pd
from pandas.core.frame import DataFrame as PandasDataFrame
from pandas.core.series import Series as PandasSeries

from dijkies.constants import BOT_STATUS, SUPPORTED_EXCHANGES
from dijkies.entities import Order, State
from dijkies.exceptions import (
    AssetNotAvailableError,
    DataTimeWindowShorterThanSuggestedAnalysisWindowError,
    InvalidExchangeAssetClientError,
    InvalidTypeForTimeColumnError,
    MissingOHLCVColumnsError,
    TimeColumnNotDefinedError,
)

logger = logging.getLogger(__name__)


class Metric(ABC):
    @property
    @abstractmethod
    def metric_name(self) -> str:
        pass

    @abstractmethod
    def calculate(self, time_series: PandasSeries) -> float:
        pass


class DataPipeline(ABC):
    @abstractmethod
    def run(self) -> PandasDataFrame:
        pass


class ExchangeMarketAPI(ABC):
    @abstractmethod
    def get_candles(
        self,
        base: str,
        interval_in_minutes: int,
        lookback_in_minutes: int,
    ) -> PandasDataFrame:
        pass

    @abstractmethod
    def get_price(self, base: str) -> float:
        pass


class ExchangeAssetClient(ABC):
    def __init__(self, state: State) -> None:
        self.state = state

    @abstractmethod
    def assets_in_state_are_available(self) -> bool:
        pass

    @abstractmethod
    def place_limit_buy_order(
        self, limit_price: float, amount_in_quote: float
    ) -> Order:
        pass

    @abstractmethod
    def place_limit_sell_order(
        self, limit_price: float, amount_in_base: float
    ) -> Order:
        pass

    @abstractmethod
    def place_market_buy_order(self, amount_in_quote: float) -> Order:
        pass

    @abstractmethod
    def place_market_sell_order(self, amount_in_base: float) -> Order:
        pass

    @abstractmethod
    def get_order_info(self, order: Order) -> Order:
        pass

    @abstractmethod
    def cancel_order(self, order: Order) -> Order:
        pass

    def update_state(self) -> None:
        for order in self.state.open_orders:
            newest_info_order = self.get_order_info(order)
            if order.is_not_equal(newest_info_order):
                self.state.process_filled_order(newest_info_order)


class Strategy(ABC):
    def __init__(
        self,
        executor: ExchangeAssetClient,
    ) -> None:
        self.executor = executor
        self.state = self.executor.state

    @abstractmethod
    def execute(self, data: PandasDataFrame) -> None:
        pass

    def run(self, data: PandasDataFrame) -> None:
        self.executor.update_state()
        if not self.executor.assets_in_state_are_available():
            raise AssetNotAvailableError(self.state.base)
        self.execute(data)

    @classmethod
    def _get_strategy_params(cls) -> list[str]:
        subclass_sig = inspect.signature(cls.__init__)
        base_sig = inspect.signature(Strategy.__init__)

        subclass_params = {
            name: p for name, p in subclass_sig.parameters.items() if name != "self"
        }
        base_params = {
            name: p for name, p in base_sig.parameters.items() if name != "self"
        }

        unique_params = {
            name: p for name, p in subclass_params.items() if name not in base_params
        }

        return list(unique_params.keys())

    def params_to_json(self):
        params = self._get_strategy_params()
        return {p: getattr(self, p) for p in params}

    def __getstate__(self):
        state = self.__dict__.copy()
        state["executor"] = None
        state["data_pipeline"] = None
        return state

    @property
    @abstractmethod
    def analysis_dataframe_size_in_minutes(self) -> int:
        pass

    def get_data_pipeline(self) -> DataPipeline:
        """
        implement this method for deployement
        """
        raise NotImplementedError()

    def backtest(self, data: PandasDataFrame) -> PandasDataFrame:
        """
        This method runs the backtest.
        It expects data, this should have the following properties:
        """

        from dijkies.executors import BacktestExchangeAssetClient
        from dijkies.performance import PerformanceInformationRow

        # validate args

        if "time" not in data.columns:
            raise TimeColumnNotDefinedError()

        if not pd.api.types.is_datetime64_any_dtype(data.time):
            raise InvalidTypeForTimeColumnError()

        lookback_in_min = self.analysis_dataframe_size_in_minutes
        timespan_data_in_min = (data.time.max() - data.time.min()).total_seconds() / 60

        if lookback_in_min > timespan_data_in_min:
            raise DataTimeWindowShorterThanSuggestedAnalysisWindowError()

        if not {"open", "high", "low", "close", "volume"}.issubset(data.columns):
            raise MissingOHLCVColumnsError()

        if not isinstance(self.executor, BacktestExchangeAssetClient):
            raise InvalidExchangeAssetClientError()

        start_time = data.iloc[0].time + timedelta(minutes=lookback_in_min)
        simulation_df: PandasDataFrame = data.loc[data.time >= start_time]
        start_candle = simulation_df.iloc[0]
        start_value_in_quote = self.state.total_value_in_quote(start_candle.open)
        result: list[PerformanceInformationRow] = []

        def get_analysis_df(
            data: PandasDataFrame, current_time: datetime, look_back_in_min: int
        ) -> PandasDataFrame:
            start_analysis_df = current_time - timedelta(minutes=look_back_in_min)

            analysis_df = data.loc[
                (data.time >= start_analysis_df) & (data.time <= current_time)
            ]

            return analysis_df.copy()

        for _, candle in simulation_df.iterrows():
            analysis_df = get_analysis_df(data, candle.time, lookback_in_min)
            self.executor.update_current_candle(candle)

            self.run(analysis_df)

            result.append(
                PerformanceInformationRow.from_objects(
                    candle, start_candle, self.state, start_value_in_quote
                )
            )

        return pd.DataFrame([r.model_dump() for r in result])


class StrategyRepository(ABC):
    @abstractmethod
    def store(
        self,
        strategy: Strategy,
        person_id: str,
        exchange: SUPPORTED_EXCHANGES,
        bot_id: str,
        status: BOT_STATUS,
    ) -> None:
        pass

    @abstractmethod
    def read(
        self,
        person_id: str,
        exchange: SUPPORTED_EXCHANGES,
        bot_id: str,
        status: BOT_STATUS,
    ) -> Strategy:
        pass

    @abstractmethod
    def change_status(
        self,
        person_id: str,
        exchange: SUPPORTED_EXCHANGES,
        bot_id: str,
        from_status: BOT_STATUS,
        to_status: BOT_STATUS,
    ) -> None:
        pass


class CredentialsRepository(ABC):
    @abstractmethod
    def get_api_key(self, person_id: str, exchange: str) -> str:
        pass

    @abstractmethod
    def get_api_secret_key(self, person_id: str, exchange: str) -> str:
        pass



class TimeColumnNotDefinedError(Exception):
    def __init__(self):
        super().__init__("the provided data should have a 'time' column")


class InvalidTypeForTimeColumnError(Exception):
    def __init__(self):
        super().__init__("'time' column has not the right dtype")


class DataTimeWindowShorterThanSuggestedAnalysisWindowError(Exception):
    def __init__(self):
        super().__init__(
            """
            the timespan of provided data is shorter than the analysis window,
            so no backtest can be executed.
            """
        )


class MissingOHLCVColumnsError(Exception):
    def __init__(self):
        super().__init__("one of the OHLCV columns is missing in the provided data")


class DataTimeSpanDifferentFromAlgorithmSetting(Exception):
    def __init__(self):
        super().__init__("one of the OHLCV columns is missing in the provided data")


class InvalidExchangeAssetClientError(Exception):
    def __init__(self):
        super().__init__("use BacktestExchangeAssetExecutor! glad that I saved you :)")


class NoOrderFoundError(Exception):
    def __init__(self, order_id):
        super().__init__(f"order with order_id {order_id} not found.")


class MultipleOrdersFoundError(Exception):
    def __init__(self, order_id):
        super().__init__(f"multiple orders found with order_id {order_id}.")


class PlaceOrderError(Exception):
    def __init__(self, message: str):
        super().__init__(f"an error occured during order creation: {message}")


class GetOrderInfoError(Exception):
    def __init__(self, message: str):
        super().__init__(f"an error occured during order info retrieval: {message}")


class CancelOrderError(Exception):
    def __init__(self, message: str):
        super().__init__(f"an error occured during cancelling order: {message}")


class MethodNotDefinedError(Exception):
    def __init__(self):
        super().__init__("method not implemented...")


class InsufficientBalanceError(Exception):
    def __init__(
        self,
        balance: dict[str, float],
        requested: float,
    ):
        super().__init__(
            f"""
            not enough balance:\n
            available: {balance["available"]}, requested: {requested}\n
            """
        )


class InsufficientOrderValueError(Exception):
    def __init__(self):
        super().__init__(
            """
            order value should be at least 5 euro:
            """
        )


class AssetNotAvailableError(Exception):
    def __init__(self, asset: str):
        super().__init__(
            f"""
            the asset {asset} in the strategy state is not available in the exchange.
            """
        )



import pandas as pd
from pandas.core.frame import DataFrame as PandasDataFrame

from dijkies.interfaces import DataPipeline, ExchangeMarketAPI


class NoDataPipeline(DataPipeline):
    def run(self) -> PandasDataFrame:
        return pd.DataFrame([{}], columns=["close"])


class CurrentValueDataPipeline(DataPipeline):
    def __init__(self, base: str, exchange_market_api: ExchangeMarketAPI) -> None:
        self.exchange_market_api = exchange_market_api
        self.base = base

    def run(self) -> PandasDataFrame:
        current_price = self.exchange_market_api.get_price(self.base)
        return pd.DataFrame([{"close": current_price}])


class OHLCVDataPipeline(DataPipeline):
    def __init__(
        self,
        exchange_market_api: ExchangeMarketAPI,
        base: str,
        candle_interval_in_minutes: int,
        lookback_in_minutes: int,
    ) -> None:
        self.exchange_market_api = exchange_market_api
        self.base = base
        self.candle_interval_in_minutes = candle_interval_in_minutes
        self.lookback_in_minutes = lookback_in_minutes

    def run(self) -> PandasDataFrame:
        return self.exchange_market_api.get_candles(
            self.base, self.candle_interval_in_minutes, self.lookback_in_minutes
        ).iloc[
            :-1
        ]  # Exclude the current forming candle



import logging
import threading
import time
from concurrent.futures import ThreadPoolExecutor, as_completed

import pandas as pd
import requests
from binance.client import Client
from pandas.core.frame import DataFrame as PandasDataFrame
from python_bitvavo_api.bitvavo import Bitvavo

from dijkies.interfaces import ExchangeMarketAPI

logger = logging.getLogger(__name__)


class BinanceMarketAPI(ExchangeMarketAPI):
    def __init__(
        self,
        logger: logging.Logger = logger,
    ):
        self.logger = logger
        self.binance_data_client = Client()

    def get_candles(
        self,
        base: str = "BTC",
        interval_in_minutes: int = 60,
        lookback_in_minutes: int = 24 * 62 * 60,
    ) -> PandasDataFrame:
        trading_pair = base + "USDT"
        interval = f"{int(interval_in_minutes / 60)}h"
        lookback = f"{lookback_in_minutes} min ago UTC"
        df = pd.DataFrame(
            self.binance_data_client.get_historical_klines(
                symbol=trading_pair, interval=interval, start_str=lookback
            )
        )

        df = df.iloc[:, :6]
        df.columns = ["time", "open", "high", "low", "close", "volume"]
        df.time = pd.to_datetime(df.time, unit="ms", utc=True)
        df[["open", "high", "low", "close", "volume"]] = df[
            ["open", "high", "low", "close", "volume"]
        ].astype(float)

        return df

    def get_price(self, base: str = "BTC") -> float:
        trading_pair = base + "USDT"
        key = f"https://api.binance.com/api/v3/ticker/price?symbol={trading_pair}"
        data = requests.get(key)
        data = data.json()
        return float(data["price"])  # type: ignore


class BitvavoMarketAPI(ExchangeMarketAPI):
    def __init__(
        self,
        logger: logging.Logger = logger,
        max_workers: int = 4,
        rate_limit_threshold: int = 50,
    ):
        self.logger = logger
        self.bitvavo_data_client = Bitvavo()
        self.max_workers = max_workers
        self.rate_limit_threshold = rate_limit_threshold
        self._lock = threading.Lock()  # shared rate-limit lock

    def _wait_if_rate_limited(self):
        """Check Bitvavo remaining limit and sleep if it's too low."""
        remaining = self.bitvavo_data_client.getRemainingLimit()
        if remaining < self.rate_limit_threshold:
            self.logger.warning(
                f"💤 Rate limit low ({remaining}). Sleeping 60s to recover."
            )
            time.sleep(60)

    def _fetch_candle_chunk(
        self, trading_pair: str, interval: str, start: int, end: int
    ) -> PandasDataFrame:
        """Fetch a single chunk of candles and return as DataFrame."""
        options = {"start": f"{start}", "end": f"{end}"}

        with self._lock:
            # Ensure only one thread checks/sleeps for rate limit at a time
            self._wait_if_rate_limited()

        try:
            candles = self.bitvavo_data_client.candles(trading_pair, interval, options)
        except Exception as e:
            self.logger.error(f"⚠️ Error fetching candles ({start}-{end}): {e}")
            return pd.DataFrame(
                [], columns=["time", "open", "high", "low", "close", "volume"]
            )

        if not candles:
            return pd.DataFrame(
                [], columns=["time", "open", "high", "low", "close", "volume"]
            )

        df = pd.DataFrame(
            candles, columns=["time", "open", "high", "low", "close", "volume"]
        )
        df.time = pd.to_datetime(df.time, unit="ms", utc=True)
        df[["open", "high", "low", "close", "volume"]] = df[
            ["open", "high", "low", "close", "volume"]
        ].astype(float)
        df = df.iloc[::-1].reset_index(drop=True)
        return df

    def get_candles(
        self,
        base: str = "BTC",
        interval_in_minutes: int = 60,
        lookback_in_minutes: int = 60 * 24 * 60,
    ) -> PandasDataFrame:
        trading_pair = base + "-EUR"

        # Determine correct Bitvavo interval
        if interval_in_minutes < 3:
            corrected_interval = 1
            interval = "1m"
        elif interval_in_minutes < 10:
            corrected_interval = 5
            interval = "5m"
        elif interval_in_minutes < 23:
            corrected_interval = 15
            interval = "15m"
        elif interval_in_minutes < 45:
            corrected_interval = 30
            interval = "30m"
        elif interval_in_minutes < 90:
            corrected_interval = 60
            interval = "1h"
        elif interval_in_minutes < 180:
            corrected_interval = 120
            interval = "2h"
        elif interval_in_minutes < 300:
            corrected_interval = 240
            interval = "4h"
        elif interval_in_minutes < 420:
            corrected_interval = 360
            interval = "6h"
        elif interval_in_minutes < 600:
            corrected_interval = 480
            interval = "8h"
        elif interval_in_minutes < 1080:
            corrected_interval = 720
            interval = "12h"
        elif interval_in_minutes < 4 * 1440:
            corrected_interval = 1440
            interval = "1d"
        elif interval_in_minutes < 19 * 1440:
            corrected_interval = 1440 * 7
            interval = "1W"
        else:
            corrected_interval = 1440 * 30
            interval = "1M"

        now = int(time.time() * 1000)
        start = now - lookback_in_minutes * 60000

        # Split into chunks
        chunk_size_minutes = (
            corrected_interval * 1440
        )  # 1440 x interval minutes per chunk
        chunks = []
        s = start
        while s < now:
            e = s + chunk_size_minutes * 60000
            chunks.append((s, e))
            s = e - corrected_interval * 60000  # small overlap to prevent gaps

        self.logger.info(
            f"""
            📮 Fetching {len(chunks)} chunks in parallel (interval={interval},
            max_workers={self.max_workers})
            """
        )

        results = []
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            future_to_range = {
                executor.submit(
                    self._fetch_candle_chunk, trading_pair, interval, s, e
                ): (s, e)
                for (s, e) in chunks
            }

            for future in as_completed(future_to_range):
                s, e = future_to_range[future]
                try:
                    df = future.result()
                    results.append(df)
                except Exception as exc:
                    self.logger.error(f"❌ Chunk {s}-{e} generated an exception: {exc}")

        if not results:
            return pd.DataFrame(
                [], columns=["time", "open", "high", "low", "close", "volume"]
            )

        all_df = (
            pd.concat(results)
            .drop_duplicates()
            .sort_values("time")
            .reset_index(drop=True)
        )
        self.logger.info(f"✅ Retrieved {len(all_df)} candles successfully.")
        return all_df

    def get_price(self, base: str = "BTC") -> float:
        trading_pair = base + "-EUR"
        price = self.bitvavo_data_client.tickerPrice({"market": trading_pair})

        return float(price["price"])



import logging
import time
import uuid
from decimal import ROUND_DOWN, Decimal, getcontext

import pandas as pd
from pandas.core.series import Series
from python_bitvavo_api.bitvavo import Bitvavo

from dijkies.constants import SUPPORTED_EXCHANGES
from dijkies.entities import Order, State
from dijkies.exceptions import (
    GetOrderInfoError,
    InsufficientBalanceError,
    InsufficientOrderValueError,
)
from dijkies.interfaces import CredentialsRepository, ExchangeAssetClient

logger = logging.getLogger(__name__)


class BacktestExchangeAssetClient(ExchangeAssetClient):
    def __init__(
        self, state: State, fee_market_order: float, fee_limit_order: float
    ) -> None:
        super().__init__(state)
        self.fee_market_order = fee_market_order
        self.fee_limit_order = fee_limit_order
        self.current_candle = pd.Series({"high": 80000, "low": 78000, "close": 79000})

    def assets_in_state_are_available(self) -> bool:
        return True

    def update_current_candle(self, current_candle: Series) -> None:
        self.current_candle = current_candle

    def place_limit_buy_order(
        self, limit_price: float, amount_in_quote: float
    ) -> Order:
        order = Order(
            order_id=str(uuid.uuid4()),
            exchange="bitvavo",
            time_created=int(time.time()),
            market=self.state.base,
            side="buy",
            limit_price=limit_price,
            on_hold=amount_in_quote,
            status="open",
            is_taker=False,
        )

        self.state.add_order(order)

        return order

    def place_limit_sell_order(
        self, limit_price: float, amount_in_base: float
    ) -> Order:
        order = Order(
            order_id=str(uuid.uuid4()),
            exchange="bitvavo",
            time_created=int(time.time()),
            market=self.state.base,
            side="sell",
            limit_price=limit_price,
            on_hold=amount_in_base,
            status="open",
            is_taker=False,
        )

        self.state.add_order(order)

        return order

    def place_market_buy_order(self, amount_in_quote: float) -> Order:
        fee = amount_in_quote * self.fee_market_order / (1 + self.fee_market_order)
        amount_in_base = (amount_in_quote - fee) / self.current_candle.close

        order = Order(
            order_id=str(uuid.uuid4()),
            exchange="bitvavo",
            time_created=int(time.time()),
            market=self.state.base,
            side="buy",
            filled=amount_in_base,
            filled_quote=amount_in_quote - fee,
            status="filled",
            fee=fee,
            is_taker=True,
        )

        self.state.process_filled_order(order)

        return order

    def place_market_sell_order(self, amount_in_base: float) -> Order:
        amount_in_quote = amount_in_base * self.current_candle.close
        fee = amount_in_quote * self.fee_market_order

        order = Order(
            order_id=str(uuid.uuid4()),
            exchange="bitvavo",
            time_created=int(time.time()),
            market=self.state.base,
            side="sell",
            filled=amount_in_base,
            filled_quote=amount_in_quote,
            status="filled",
            fee=fee,
            is_taker=True,
        )

        self.state.process_filled_order(order)

        return order

    def get_order_info(self, order: Order) -> Order:
        found_order = self.state.get_order(order.order_id)
        if found_order.status == "open":
            is_filled = (
                found_order.side == "buy"
                and found_order.limit_price >= self.current_candle.low
            ) or (
                found_order.side == "sell"
                and found_order.limit_price <= self.current_candle.high
            )
            if is_filled:
                return self.fill_open_order(found_order)
        return found_order

    def cancel_order(self, order: Order) -> Order:
        self.state.cancel_order(order)
        return order

    def fill_open_order(self, order: Order) -> Order:
        fee_limit_order = self.fee_limit_order
        if order.status != "open":
            raise ValueError("only open orders can be filled")
        if order.side == "buy":
            fee = order.on_hold * fee_limit_order / (1 + fee_limit_order)
            filled_quote = order.on_hold - fee
            filled = filled_quote / order.limit_price  # type: ignore
        else:
            filled = order.on_hold
            filled_quote = order.on_hold * order.limit_price  # type: ignore
            fee = filled_quote * fee_limit_order
        return Order(
            order_id=order.order_id,
            exchange=order.exchange,
            time_created=order.time_created,
            market=order.market,
            side=order.side,
            limit_price=order.limit_price,
            on_hold=0,
            status="filled",
            is_taker=False,
            fee=fee,
            filled_quote=filled_quote,
            filled=filled,
        )


def order_from_bitvavo_response(response: dict) -> Order:
    return Order(
        exchange="bitvavo",
        order_id=response["orderId"],
        market=response["market"],
        time_created=int(response["created"]),
        time_canceled=None,
        time_filled=(
            max([int(fill["timestamp"]) for fill in response["fills"]])
            if len(response["fills"]) > 0
            else None
        ),
        on_hold=float(response["onHold"]),
        side=response["side"],
        limit_price=float(response["price"]) if "price" in response else None,
        actual_price=(
            float(response["filledAmountQuote"]) / float(response["filledAmount"])
            if float(response["filledAmount"]) > 0
            else None
        ),
        filled=float(response["filledAmount"]),
        filled_quote=float(response["filledAmountQuote"]),
        fee=float(response["feePaid"]),
        is_taker=response["fills"][0]["taker"] if response["fills"] else False,
        status=(
            response["status"]
            if response["status"] in ["filled", "cancelled"]
            else "open"
        ),
    )


class BitvavoExchangeAssetClient(ExchangeAssetClient):
    max_fee = 0.0025

    def __init__(
        self,
        state: State,
        bitvavo_api_key: str,
        bitvavo_api_secret_key: str,
        operator_id: int,
    ) -> None:
        super().__init__(state)
        self.operator_id = operator_id
        self.bitvavo = Bitvavo(
            {
                "APIKEY": bitvavo_api_key,
                "APISECRET": bitvavo_api_secret_key,
                "RESTURL": "https://api.bitvavo.com/v2",
                "WSURL": "wss://ws.bitvavo.com/v2/",
                "ACCESSWINDOW": 10000,
                "DEBUGGING": False,
            }
        )

    def assets_in_state_are_available(self) -> bool:
        base_response = self.bitvavo.balance({"symbol": self.state.base})
        quote_response = self.bitvavo.balance({"symbol": "EUR"})

        available_base = float(base_response[0]["available"]) if base_response else 0
        available_quote = float(quote_response[0]["available"]) if quote_response else 0
        in_order_base = float(base_response[0]["inOrder"]) if base_response else 0
        in_order_quote = float(quote_response[0]["inOrder"]) if quote_response else 0

        base_is_available = self.state.base_available <= available_base * 1.000001
        quote_is_available = self.state.quote_available <= available_quote * 1.000001
        in_order_base_is_available = self.state.base_on_hold <= in_order_base * 1.000001
        in_order_quote_is_available = (
            self.state.quote_on_hold <= in_order_quote * 1.000001
        )

        logger.info(
            "base available exchange: %s | base available state: %s",
            available_base,
            self.state.base_available,
        )
        logger.info(
            "quote available exchange: %s | quote available state: %s",
            available_quote,
            self.state.quote_available,
        )
        logger.info(
            "base in order exchange: %s | base on hold state: %s",
            in_order_base,
            self.state.base_on_hold,
        )
        logger.info(
            "quote in order exchange: %s | quote on hold state: %s",
            in_order_quote,
            self.state.quote_on_hold,
        )

        return (
            base_is_available
            and quote_is_available
            and in_order_base_is_available
            and in_order_quote_is_available
        )

    def quantity_decimals(self) -> int:
        trading_pair = self.state.base + "-EUR"
        return self.bitvavo.markets({"market": trading_pair})["quantityDecimals"]

    @staticmethod
    def __closest_valid_price(price: float) -> float:
        getcontext().prec = 20
        price = Decimal(str(price))
        x = 0

        ten = Decimal("10")

        if price > 1:
            while price / (ten**x) > 1:
                x += 1
        else:
            while price / (ten**x) < 1:
                x -= 1
            x += 1

        shifted = price / (ten**x)
        rounded = shifted.quantize(Decimal("1.00000"), rounding=ROUND_DOWN)
        corrected = rounded * (ten**x)

        return float(corrected)

    def get_balance_base(self) -> dict[str, float]:
        balance = self.bitvavo.balance({"symbol": self.state.base})
        logger.info(f"response Balance base: {balance}")
        if balance:
            balance = balance[0]
        else:
            balance = {"available": 0, "inOrder": 0}
        return balance

    def get_balance_quote(self) -> dict[str, float]:
        balance = self.bitvavo.balance({"symbol": "EUR"})
        logger.info(f"response Balance quote: {balance}")
        if balance:
            balance = balance[0]
        else:
            balance = {"available": 0, "inOrder": 0}
        return balance

    def place_limit_buy_order(
        self, limit_price: float, amount_in_quote: float
    ) -> Order:
        trading_pair = self.state.base + "-EUR"
        limit_price = self.__closest_valid_price(price=float(limit_price))

        amount_in_base = round(
            (float(amount_in_quote) - 0.01) / (limit_price * (1 + self.max_fee)),
            self.quantity_decimals(),
        )
        logger.info(
            f"""
            place limit buy order;
            market: {trading_pair}
            limitPrice: {limit_price}
            amount: {amount_in_base}
            operatorId: {self.operator_id}
"""
        )
        response = self.bitvavo.placeOrder(
            market=trading_pair,
            side="buy",
            orderType="limit",
            body={
                "amount": str(amount_in_base),
                "price": str(limit_price),
                "operatorId": self.operator_id,
            },
        )
        logger.info(f"place limit buy order response: {response}")
        if "errorCode" in response:
            error_code = response["errorCode"]
            if error_code in [107, 108, 109]:
                time.sleep(3)
                order = self.place_limit_buy_order(limit_price, amount_in_quote)
                return order
            elif error_code == 216:
                balance = self.get_balance_quote()
                raise InsufficientBalanceError(balance, amount_in_quote)
            elif error_code == 217:
                raise InsufficientOrderValueError()
            else:
                raise Exception(str(response))

        order = order_from_bitvavo_response(response)
        if order.is_filled:
            """
            if the order is immediately filled, we have to treat the order as a market order.
            wait a bit to make sure the fills are registered in the exchange
            before fetching and processing the order info
            """
            time.sleep(3)
            order = self.get_order_info(order)
            self.state.process_filled_order(order)
        else:
            self.state.add_order(order)
        return order

    def place_limit_sell_order(
        self, limit_price: float, amount_in_base: float
    ) -> Order:
        trading_pair = self.state.base + "-EUR"

        quantity_decimals = self.quantity_decimals()
        factor = 1 / (10**quantity_decimals)
        amount_in_base = round(
            (float(amount_in_base) // factor) * factor,
            quantity_decimals,
        )

        limit_price = self.__closest_valid_price(price=float(limit_price))
        logger.info(
            f"""
            place limit sell order
            market: {trading_pair}
            amount: {amount_in_base}
            price: {limit_price}
            operatorId: {self.operator_id}
            """
        )
        response = self.bitvavo.placeOrder(
            market=trading_pair,
            side="sell",
            orderType="limit",
            body={
                "amount": str(amount_in_base),
                "price": str(limit_price),
                "operatorId": self.operator_id,
            },
        )
        logger.info(f"place limit sell order response: {response}")
        if "errorCode" in response:
            error_code = response["errorCode"]
            if error_code in [107, 108, 109]:
                time.sleep(3)
                order = self.place_limit_sell_order(limit_price, amount_in_base)
                return order
            elif error_code == 216:
                balance = self.get_balance_base()
                raise InsufficientBalanceError(balance, amount_in_base)
            elif error_code == 217:
                raise InsufficientOrderValueError()
            else:
                raise Exception(str(response))

        order = order_from_bitvavo_response(response)
        if order.is_filled:
            """
            if the order is immediately filled, we have to treat the order as a market order.
            wait a bit to make sure the fills are registered in the exchange
            before fetching and processing the order info
            """
            time.sleep(3)
            order = self.get_order_info(order)
            self.state.process_filled_order(order)
        else:
            self.state.add_order(order)
        return order

    def place_market_buy_order(self, amount_in_quote: float) -> Order:
        trading_pair = self.state.base + "-EUR"

        amount_in_quote = str(round(float(amount_in_quote), 2))
        logger.info(
            f"""
            place market buy order
            market: {trading_pair}
            amountQuote: {amount_in_quote}
            operatorId: {self.operator_id}
            """
        )
        response = self.bitvavo.placeOrder(
            market=trading_pair,
            side="buy",
            orderType="market",
            body={"amountQuote": amount_in_quote, "operatorId": self.operator_id},
        )
        logger.info(f"place market buy order response: {response}")
        if "errorCode" in response:
            error_code = response["errorCode"]
            if error_code in [107, 108, 109]:
                time.sleep(3)
                order = self.place_market_buy_order(amount_in_quote)
                return order
            elif error_code == 216:
                balance = self.get_balance_quote()
                raise InsufficientBalanceError(balance, amount_in_quote)
            elif error_code == 217:
                raise InsufficientOrderValueError()
            else:
                raise Exception(str(response))

        order = order_from_bitvavo_response(response)
        # wait a bit to make sure the fills are registered in the exchange
        time.sleep(3)
        order = self.get_order_info(order)
        self.state.process_filled_order(order)
        return order

    def place_market_sell_order(self, amount_in_base: float) -> Order:
        trading_pair = self.state.base + "-EUR"
        quantity_decimals = self.quantity_decimals()

        factor = 1 / (10**quantity_decimals)
        amount_in_base = round(
            (float(amount_in_base) // factor) * factor,
            quantity_decimals,
        )
        logger.info(
            f"""
            place market sell order
            market: {trading_pair}
            amount: {amount_in_base}
            operatorId: {self.operator_id}
            """
        )
        response = self.bitvavo.placeOrder(
            market=trading_pair,
            side="sell",
            orderType="market",
            body={"amount": str(amount_in_base), "operatorId": self.operator_id},
        )
        logger.info(f"place market sell order response: {response}")
        if "errorCode" in response:
            error_code = response["errorCode"]
            if error_code in [107, 108, 109]:
                time.sleep(3)
                order = self.place_market_sell_order(amount_in_base)
                return order
            elif error_code == 216:
                balance = self.get_balance_base()
                raise InsufficientBalanceError(balance, amount_in_base)
            elif error_code == 217:
                raise InsufficientOrderValueError()
            else:
                raise Exception(str(response))

        order = order_from_bitvavo_response(response)
        # wait a bit to make sure the fills are registered in the exchange
        time.sleep(3)
        order = self.get_order_info(order)
        self.state.process_filled_order(order)
        return order

    def get_order_info(self, order: Order) -> Order:

        response = self.bitvavo.getOrder(market=order.market, orderId=order.order_id)
        logger.info(f"get order info response: {response}")
        if "errorCode" in response:
            if response["errorCode"] == 240:
                raise GetOrderInfoError(response)
        return order_from_bitvavo_response(response)

    def cancel_order(self, order: Order) -> Order:
        response = self.bitvavo.cancelOrder(
            market=order.market,
            orderId=order.order_id,
            operatorId=self.operator_id,
        )
        logger.info(f"cancel order response: {response}")
        if "errorCode" in response:
            if response["errorCode"] != 240:
                raise GetOrderInfoError(response)
        self.state.cancel_order(order)
        return order


def get_executor(
    person_id: str,
    exchange: SUPPORTED_EXCHANGES,
    state: State,
    credentials_repository: CredentialsRepository,
) -> ExchangeAssetClient:
    if exchange == "bitvavo":
        api_key = credentials_repository.get_api_key(person_id, exchange)
        api_secret_key = credentials_repository.get_api_secret_key(person_id, exchange)
        return BitvavoExchangeAssetClient(state, api_key, api_secret_key, operator_id=1)
    elif exchange == "backtest":
        return BacktestExchangeAssetClient(
            state, fee_market_order=0.0025, fee_limit_order=0.0015
        )
    else:
        raise ValueError(f"Unsupported exchange: {exchange}")



from typing import Literal, Optional, Union

from pydantic import BaseModel

from dijkies.constants import SUPPORTED_EXCHANGES
from dijkies.exceptions import (
    MultipleOrdersFoundError,
    NoOrderFoundError,
)


class Order(BaseModel):
    order_id: str
    exchange: SUPPORTED_EXCHANGES
    market: str
    time_created: int
    time_canceled: Union[int, None] = None
    time_filled: Union[int, None] = None
    on_hold: float = 0
    side: Literal["buy", "sell"]
    limit_price: Optional[float] = None
    actual_price: Optional[float] = None
    filled: float = 0
    filled_quote: float = 0
    fee: float = 0
    is_taker: bool
    status: Literal["open", "filled", "cancelled"]

    @property
    def is_filled(self) -> bool:
        return self.status == "filled"

    @property
    def is_open(self) -> bool:
        return self.status == "open"

    @property
    def is_cancelled(self) -> bool:
        return self.status == "cancelled"

    def is_equal(self, order: "Order") -> bool:
        return self.status == order.status

    def is_not_equal(self, order: "Order") -> bool:
        return not self.is_equal(order)


class State(BaseModel):
    base: str
    total_base: float
    total_quote: float
    orders: list[Order] = []

    @property
    def number_of_transactions(self) -> int:
        return len(self.filled_orders)

    @property
    def total_fee_paid(self) -> float:
        return sum([o.fee for o in self.filled_orders])

    @property
    def filled_orders(self) -> list[Order]:
        return [o for o in self.orders if o.is_filled]

    @property
    def open_orders(self) -> list[Order]:
        return [o for o in self.orders if o.is_open]

    @property
    def cancelled_orders(self) -> list[Order]:
        return [o for o in self.orders if o.is_cancelled]

    @property
    def base_on_hold(self) -> float:
        return sum([order.on_hold for order in self.sell_orders])

    @property
    def quote_on_hold(self) -> float:
        return sum([order.on_hold for order in self.buy_orders])

    @property
    def base_available(self) -> float:
        return self.total_base - self.base_on_hold

    @property
    def quote_available(self) -> float:
        return self.total_quote - self.quote_on_hold

    @property
    def buy_orders(self) -> list[Order]:
        return [o for o in self.open_orders if o.side == "buy"]

    @property
    def sell_orders(self) -> list[Order]:
        return [o for o in self.open_orders if o.side == "sell"]

    def add_order(self, order: Order) -> None:
        self.orders.append(order)

    def get_order(self, order_id: str) -> Order:
        list_found_order = [o for o in self.orders if o.order_id == order_id]
        if len(list_found_order) == 0:
            raise NoOrderFoundError(order_id)
        elif len(list_found_order) > 1:
            raise MultipleOrdersFoundError(order_id)
        return list_found_order[0]

    def cancel_order(self, order: Order) -> None:
        found_order = self.get_order(order.order_id)
        found_order.status = "cancelled"

    def process_filled_order(self, filled_order: Order) -> None:
        if filled_order.side == "buy":
            quote_mutation = -(filled_order.filled_quote + filled_order.fee)
            base_mutation = filled_order.filled
        else:
            quote_mutation = filled_order.filled_quote - filled_order.fee
            base_mutation = -filled_order.filled

        self.total_quote += quote_mutation
        self.total_base += base_mutation

        if filled_order.is_taker:
            self.add_order(filled_order)
        else:
            found_order = self.get_order(filled_order.order_id)
            found_order.status = "filled"
            found_order.fee = filled_order.fee

        self._check_non_negative()

    def _check_non_negative(self) -> None:
        if self.base_available < -1e-9:
            raise ValueError(f"Negative base balance: {self.base_available}")
        if self.quote_available < -1e-9:
            raise ValueError(f"Negative quote balance: {self.quote_available}")

    def total_value_in_base(self, price: float) -> float:
        return self.total_base + self.total_quote / price

    def total_value_in_quote(self, price: float) -> float:
        return self.total_quote + self.total_base * price

    def fraction_value_in_quote(self, price: float) -> float:
        return self.total_quote / max(self.total_value_in_quote(price), 0.00000001)

    def fraction_value_in_base(self, price: float) -> float:
        return 1 - self.fraction_value_in_quote(price)



import uuid
from datetime import datetime

import numpy as np
from pandas.core.series import Series
from pandas.core.series import Series as PandasSeries
from pydantic import BaseModel

from dijkies.entities import State
from dijkies.interfaces import Metric


class PerformanceInformationRow(BaseModel):
    id: str = str(uuid.uuid4())
    candle_time: datetime
    candle_open: float
    candle_high: float
    candle_low: float
    candle_close: float
    buy_orders: list = []
    sell_orders: list = []
    balance_total_base: float
    balance_total_quote: float
    balance_available_base: float
    balance_available_quote: float
    balance_base_on_hold: float
    balance_quote_on_hold: float
    total_fee_paid: float
    total_value_strategy: float
    roi_strategy: float
    roi_strategy_per_month: float
    total_value_hodl: float
    roi_hodl: float
    roi_hodl_per_month: float
    number_of_transactions: int
    absolute_profit: float

    @classmethod
    def from_objects(
        cls,
        candle: Series,
        start_candle: Series,
        state: State,
        initialization_value_in_quote: float,
    ) -> "PerformanceInformationRow":
        strategy_value = state.total_value_in_quote(candle.close)
        hodl_value = (candle.close / start_candle.open) * initialization_value_in_quote

        roi_strategy = ((strategy_value / initialization_value_in_quote) - 1) * 100
        roi_hodl = ((candle.close / start_candle.open) - 1) * 100

        duration_in_months = (candle.time - start_candle.time).total_seconds() / (
            60 * 60 * 24 * 30
        )

        strategy_per_month = (roi_strategy / 100 + 1) ** (
            1 / max(duration_in_months, 0.01)
        )
        hodl_per_month = (roi_hodl / 100 + 1) ** (1 / max(duration_in_months, 0.01))

        return cls(
            candle_time=candle.time,
            candle_open=candle.open,
            candle_high=candle.high,
            candle_low=candle.low,
            candle_close=candle.close,
            buy_orders=state.buy_orders,
            sell_orders=state.sell_orders,
            balance_total_base=state.total_base,
            balance_total_quote=state.total_quote,
            balance_available_base=state.base_available,
            balance_available_quote=state.quote_available,
            balance_base_on_hold=state.base_on_hold,
            balance_quote_on_hold=state.quote_on_hold,
            total_fee_paid=state.total_fee_paid,
            total_value_strategy=strategy_value,
            roi_strategy=roi_strategy,
            roi_strategy_per_month=strategy_per_month,
            total_value_hodl=hodl_value,
            roi_hodl=roi_hodl,
            roi_hodl_per_month=hodl_per_month,
            number_of_transactions=state.number_of_transactions,
            absolute_profit=strategy_value - initialization_value_in_quote,
        )


class DrawDown(Metric):
    @property
    def metric_name(self) -> str:
        return "draw_down"

    def calculate(self, time_series: PandasSeries) -> float:
        returns = time_series.pct_change()
        returns.fillna(0.0, inplace=True)

        cumulative = (returns + 1).cumprod()
        running_max = np.maximum.accumulate(cumulative)

        result = ((cumulative - running_max) / running_max).min() * 100

        return result


class ReturnOnInvestment(Metric):
    @property
    def metric_name(self) -> str:
        return "roi"

    def calculate(self, time_series: PandasSeries) -> float:
        return ((time_series.iloc[-1] / time_series.iloc[0]) - 1) * 100


class NormalizedReturnOnInvestment(Metric):
    def __init__(self, candle_interval_in_minutes: int) -> None:
        self.candle_interval_in_minutes = candle_interval_in_minutes

    @property
    def metric_name(self) -> str:
        return "normalized_roi"

    def calculate(self, time_series: PandasSeries) -> float:
        """
        returns the average return per month.
        """
        total_time_in_minutes = len(time_series) * self.candle_interval_in_minutes
        total_time_in_months = total_time_in_minutes / (60 * 24 * 30)

        total_return = time_series.iloc[-1] / time_series.iloc[0]
        return_per_month = total_return ** (min(1 / total_time_in_months, 1))

        return (return_per_month - 1) * 100


class SharpeRatio(Metric):
    def __init__(
        self, risk_free_rate_per_year: float, candle_interval_in_minutes: int
    ) -> None:
        self.risk_free_rate_per_year = risk_free_rate_per_year
        self.candle_interval_in_minutes = candle_interval_in_minutes

    @property
    def metric_name(self) -> str:
        return "sharpe_ratio"

    def calculate(self, time_series: PandasSeries) -> float:
        risk_free_rate_per_day = (1 + self.risk_free_rate_per_year) ** (1 / 365) - 1
        measurements_per_day = 60 * 24 / self.candle_interval_in_minutes
        normalized_std_return = risk_free_rate_per_day / max(measurements_per_day, 1)
        returns = time_series.pct_change()
        excess_return = returns.mean() - normalized_std_return
        volatility = max(returns.std(), 0.00001)
        sharpe_ratio = excess_return / volatility
        return sharpe_ratio * np.sqrt(measurements_per_day)



import logging
import os
import pickle
import shutil
from pathlib import Path

from dijkies.constants import ASSET_HANDLING, BOT_STATUS, SUPPORTED_EXCHANGES
from dijkies.interfaces import (
    CredentialsRepository,
    Strategy,
    StrategyRepository,
)

logger = logging.getLogger(__name__)


class LocalCredentialsRepository(CredentialsRepository):
    def get_api_key(self, person_id: str, exchange: str) -> str:
        return os.environ.get(f"{person_id}_{exchange}_api_key")

    def get_api_secret_key(self, person_id: str, exchange: str) -> str:
        return os.environ.get(f"{person_id}_{exchange}_api_secret_key")


class LocalStrategyRepository(StrategyRepository):
    def __init__(self, root_directory: Path) -> None:
        self.root_directory = root_directory

    def store(
        self,
        strategy: Strategy,
        person_id: str,
        exchange: SUPPORTED_EXCHANGES,
        bot_id: str,
        status: BOT_STATUS,
    ) -> None:
        (self.root_directory / person_id / exchange / status).mkdir(
            parents=True, exist_ok=True
        )
        path = os.path.join(
            self.root_directory, person_id, exchange, status, bot_id + ".pkl"
        )
        with open(path, "wb") as file:
            pickle.dump(strategy, file)

    def read(
        self,
        person_id: str,
        exchange: SUPPORTED_EXCHANGES,
        bot_id: str,
        status: BOT_STATUS,
    ) -> Strategy:
        path = os.path.join(
            self.root_directory, person_id, exchange, status, bot_id + ".pkl"
        )
        with open(path, "rb") as file:
            strategy = pickle.load(file)
        return strategy

    def change_status(
        self,
        person_id: str,
        exchange: SUPPORTED_EXCHANGES,
        bot_id: str,
        from_status: BOT_STATUS,
        to_status: BOT_STATUS,
    ) -> None:
        if from_status == to_status:
            return
        src = (
            Path(f"{self.root_directory}/{person_id}/{exchange}/{from_status}")
            / f"{bot_id}.pkl"
        )
        dest_folder = Path(f"{self.root_directory}/{person_id}/{exchange}/{to_status}")

        dest_folder.mkdir(parents=True, exist_ok=True)
        shutil.move(src, dest_folder / src.name)


class Bot:
    def __init__(
        self,
        strategy_repository: StrategyRepository,
        credential_repository: CredentialsRepository,
    ) -> None:
        self.strategy_repository = strategy_repository
        self.credential_repository = credential_repository

    def load_strategy(
        self,
        person_id: str,
        exchange: SUPPORTED_EXCHANGES,
        bot_id: str,
        status: BOT_STATUS,
    ) -> Strategy:
        from dijkies.executors import get_executor

        strategy = self.strategy_repository.read(person_id, exchange, bot_id, status)
        strategy.executor = get_executor(
            person_id, exchange, strategy.state, self.credential_repository
        )
        return strategy

    def run(
        self,
        person_id: str,
        exchange: SUPPORTED_EXCHANGES,
        bot_id: str,
        status: BOT_STATUS,
    ) -> None:

        strategy = self.load_strategy(person_id, exchange, bot_id, status)
        data_pipeline = strategy.get_data_pipeline()
        data = data_pipeline.run()

        try:
            strategy.run(data)
            self.strategy_repository.store(
                strategy, person_id, exchange, bot_id, status
            )
        except Exception as e:
            self.strategy_repository.store(
                strategy, person_id, exchange, bot_id, status
            )
            self.strategy_repository.change_status(
                person_id, exchange, bot_id, status, "paused"
            )
            raise Exception(e)

    def stop(
        self,
        person_id: str,
        exchange: SUPPORTED_EXCHANGES,
        bot_id: str,
        status: BOT_STATUS,
        asset_handling: ASSET_HANDLING,
    ) -> None:
        strategy = self.load_strategy(person_id, exchange, bot_id, status)

        try:
            for open_order in strategy.state.open_orders:
                _ = strategy.executor.cancel_order(open_order)
            if asset_handling == "base_only":
                _ = strategy.executor.place_market_buy_order(
                    strategy.state.quote_available
                )
            elif asset_handling == "quote_only":
                _ = strategy.executor.place_market_sell_order(
                    strategy.state.base_available
                )
            self.strategy_repository.store(
                strategy, person_id, exchange, bot_id, status
            )
            self.strategy_repository.change_status(
                person_id, exchange, bot_id, status, "stopped"
            )

        except Exception as e:
            self.strategy_repository.store(
                strategy, person_id, exchange, bot_id, status
            )
            self.strategy_repository.change_status(
                person_id, exchange, bot_id, status, "paused"
            )
            raise Exception(e)
